require 'uu/os/qos/qos_handler'
require 'uu/os/rest'
require 'uu/os/uesuri'
require 'uu/os/group/group_create'
require 'uu/os/group/group_attributes'
require 'uu/os/group/group_set_attributes'
require 'uu/os/group/group_propagate_activities'
require 'uu/os/group/group_cancel_activities'
require 'uu/os/group/group_add_role_interface'
require 'uu/os/group/group_remove_role_interface'
require 'uu/os/group/group_get_role_interface_list'
require 'uu/os/group/group_get_role_interface_list_item'



module UU
  module OS

      module Group
        extend self

        # Service path
        PATH = 'ues/core/role/UESGroup'

        # Creates a new group role. The command creates a new group in a specified location.
        # At the very least meta artifact must be specified in {GroupCreate} DTO.
        # Competent role for the new group will be selected as the most suitable according
        # to specified container and meta artifact (executive/authorized role with connected
        # interface), or can be also specified in {GroupCreate} DTO.
        #
        # @param [String, UU::OS::UESURI] location_uri UESURI of the folder/org. unit/meta model where the group will be created.
        # @param [UU::OS::Group::GroupCreate] group Specific options.
        # @option group [String, UU::OS::UESURI] :meta_artifact_uri UESURI of the meta artifact. Only a meta artifact with the template class of group (not its descendants) can be used. If the rights does not permit to create a new group from the meta artifact in the specified location, the create command fails.
        # @option group [String] :name Name of the new group. If nil is set, tName from meta artifact is used if set, otherwise the create command fails.
        # @option group [String] :code Code of the new group. If nil is set, the code is generated automatically using the system sequence. The code must be unique in the given territory, otherwise the create command fails.
        # @option group [String] :description Description of the new group. If nil is set, tDescription from meta artifact is used if set, otherwise no description is set on the new group.
        # @option group [String, UU::OS::UESURI] :competent_role_uri UESURI of the competent role. If nil is set, the executive role from authorization is used. If no role is authorized to create the group into specified container, the create command fails. If the selected role doesn't have connected the role interface with the selected meta artifact, the create command fails. If the selected role doesn't belong to the same organizational unit as the new group role or if it doesn't belong to one of superior organizational units, the create command fails.
        # @option group [String, UU::OS::Artifact::SecurityLevel] :security_level Security level. If nil is set, security level is taken from the meta artifact tSecurityLevel. If the meta artifact tSecurityLevel is higher than the user's security clearance, the create command fails.
        # @option group [Boolean] :searchable_in_territory Specifies if it is possible to search the new group in its territory. If nil is set, tSearchableInTerritory from meta artifact is used if set, otherwise the create command fails.
        # @option group [String, UU::OS::UESURI] :icon_uri URI of the group's icon. If nil is set, the group uses icon from meta artifact. An example of icon URI: 'ues:#{{}system}:#{{}ues_v5.core_v1.codetable_v1.base_v1.CodeTable_Icons}:#{{}ART_077}'.
        # @option group [Boolean] :grant_all_from_organizational_unit <b>Deprecated.</b> Specifies if the group will be included among All roles from organizational unit. If true or nil is set, the role will be included, otherwise it will not be included among All roles from organizational unit.
        # @option group [Boolean] :dry_run Specifies whether this is only a test run to verify that the object can be created. The default value is false.
        # @option group [Boolean] :component_content Defines whether the primarily displayed content of the created Group is component content. Default value is false.
        #
        # @return [UU::OS::UESURI] UESURI of the new group 
        #
        # @example
        #   # Create a new group.
        #   UU::OS::Group.create('ues:TERRITORY:FOLDER', :meta_artifact_uri => 'ues:TERRITORY:META_ARTIFACT', :code => 'NEW_GROUP', :name => 'New Group')
        #
        #   # Verifying that a new group can be created.
        #   UU::OS::Group.create('ues:TERRITORY:FOLDER', :meta_artifact_uri => 'ues:TERRITORY:META_ARTIFACT',:code => 'NEW_GROUP', :name => 'New Group', :dry_run => true)
        def create(location_uri, group = nil)
          svc = UU::OS::REST::RemoteClient.new(Group)
          payload = UU::OS::Group::GroupCreate.new(group).to_json

          UU::OS::QoS::QoSHandler.auto_retry do
            res = svc.post('create', location_uri, payload)
            # res is nil if dry_run was set to true. In that case, return nil (do nothing)
            if !(res.nil? || res.empty?)
              return UU::OS::UESURI.new(res)
            end
          end
        end

        # Returns attributes representing the group role specified by the
        # group_uri parameter. The command does not change the state of the group in the
        # system. 
        #
        # @param [String, UU::OS::UESURI] group_uri UESURI of a group whose attributes are returned 
        #
        # @return [GroupAttributes] DTO containing the group attributes
        #
        # @example
        #   # Get group attributes
        #   UU::OS::Group.get_attributes('ues:TERRITORY:GROUP')
        def get_attributes(group_uri)
          svc = UU::OS::REST::RemoteClient.new(Group)
          UU::OS::QoS::QoSHandler.auto_retry do
            res = svc.get('getAttributes', group_uri)
            return Group::GroupAttributes.new(res)
          end
        end

        # Command for setting attributes of a group role. The command cannot modify references
        # to objects like location, competent role and so on.
        #
        # @param [String, UU::OS::UESURI] group_uri UESURI of the group whose attributes are to be changed
        # @param [GroupSetAttributes] group DTO containing new attributes the group
        # @option group [String] :name The new name of the group. If nil is set, the name is not modified.
        # @option group [String] :code The new code of the group. If nil is set, the code is not modified. The code must be unique in a given territory, otherwise the set_attributes command fails.
        # @option group [String] :description The new description of the group. If nil is set, the description is not modified.
        # @option group [String, UU::OS::Artifact::SecurityLevel] :security_level The new security level of the group. If nil is set, the security level is not modified. If the meta artifact tSecurityLevel is higher than the user's security clearance, the set_attributes command fails.
        # @option group [Boolean] :searchable_in_territory Specifies if it is possible to search the group in its territory. If nil is set, this attribute on the group is not modified.
        # @option group [String, UU::OS::UESURI] :icon_uri URI of the group's icon. If nil is set, the icon is not modified. If nil URI ("ues:{[-1]}:{[-1]}:") is set, the group icon is removed and the group uses icon from meta artifact. Allowed format of icon URI: 'ues:#{{}system}:#{{}ues_v5.core_v1.codetable_v1.base_v1.CodeTable_Icons}:#{{}CODE_OF_ICON}'. An example of icon URI: 'ues:#{{}system}:#{{}ues_v5.core_v1.codetable_v1.base_v1.CodeTable_Icons}:#{{}ART_077}'.
        # @option group [Boolean] :grant_all_from_organizational_unit <b>Deprecated.</b> Specifies if the group will be included among All roles from organizational unit. If true or nil is set, the role will be included, otherwise it will not be included among All roles from organizational unit.
        # @option group [Boolean] :component_content Defines whether the primarily displayed content of the created Group is component content. Default value is false.
        #
        # @return [UU::OS::UESURI] UESURI of the group
        #
        # @example
        #   # Set attributes of a group - make it non searchable in territory.
        #   UU::OS::Group.set_attributes('ues:TERRITORY:GROUP',
        #         :name => 'John',
        #         :surname => 'Doe',
        #         :code => 'NEWCODE',
        #         :description => 'Some sample description',
        #         :searchable_in_territory => false
        #   )
        def set_attributes(group_uri, group = nil)
          svc = UU::OS::REST::RemoteClient.new(Group)
          payload = UU::OS::Group::GroupSetAttributes.new(group).to_json

          UU::OS::QoS::QoSHandler.auto_retry do
            res = svc.post('setAttributes', group_uri, payload)
            return UU::OS::UESURI.new(res)
          end
        end

        # Deletes the specified group role from the system. The group is deleted
        # even if it contains a locked sheet or attachment. The group cannot be deleted if:
        # the group is not in an active cast,
        # the group's workflow is not created,
        # a property contains a reference to this group,
        # a time sheet exists with a reference to this group,
        # an activity exists with this role as competent role.
        #
        # @param [String, UU::OS::UESURI] group_uri UESURI of the group which is about to be deleted.
        #
        # @example
        #   # Delete group specified by UESURI.
        #   UU::OS::Group.delete('ues:TERRITORY:GROUP')
        def delete(group_uri)
          svc = UU::OS::REST::RemoteClient.new(Group)
          UU::OS::QoS::QoSHandler.auto_retry do
            svc.post('delete', group_uri)
          end
        end

        # Propagates activities of the group to its casted roles or casted groups.
        # Only the active activities are propagated, it means that they have a state and the
        # state must not be final.
        #
        # @param [String, UU::OS::UESURI] group_uri UESURI of the group whose activities will be propagated
        # @param [GroupPropagateActivities] group UESURIs of the roles which must be casted into group with +group_uri+ and into which the activities will be propagated.
        # @option group [Array] :casted_roles_uris The casted roles to that the group role's activities will be propagated. Cannot be nil or empty.
        #
        # @example
        #   # Propagate activities to selected roles casted to GROUP
        #   UU::OS::Group.propagate_activities('ues:TERRITORY:GROUP',
        #         :casted_roles_uris => ['ues:CMD-BT:ROLE1', 'ues:CMD-BT:ROLE2']
        #   )
        def propagate_activities(group_uri, group = nil)
          svc = UU::OS::REST::RemoteClient.new(Group)
          payload = UU::OS::Group::GroupPropagateActivities.new(group).to_json

          UU::OS::QoS::QoSHandler.auto_retry do
            svc.post('propagateActivities', group_uri, payload)
          end
        end

        # Cancels activities which exist due to group roles for the specified roles. 
        #
        # @param [String, UU::OS::UESURI] group_role_uri UESURI of the group role with activities 
        # @param [GroupCancelActivities] group DTO containing attributes to cancel activities for casted roles
        # @option group [Array] :casted_roles_uris The casted roles to that the group role's activities will be propagated. Cannot be nil or empty.
        #
        # @example
        #   # Cancel activities of selected roles casted to GROUP
        #   UU::OS::Group.cancel_activities('ues:TERRITORY:GROUP',
        #         :casted_roles_uris => ['ues:CMD-BT:ROLE1', 'ues:CMD-BT:ROLE2']
        #   )
        def cancel_activities(group_role_uri, group = nil)
          svc = UU::OS::REST::RemoteClient.new(Group)
          payload = UU::OS::Group::GroupCancelActivities.new(group).to_json

          UU::OS::QoS::QoSHandler.auto_retry do
            svc.post('cancelActivities', group_role_uri, payload)
          end
        end

        # Creates new role interface connection to the specified role. The
        # role interface has to be in an active state and must not be already connected. The
        # role interface has to be connected to the one of the authorized roles for this command,
        # or the authorized role has to be casted to the default group of meta model where the
        # role interface is located.
        #
        # @param [String, UU::OS::UESURI] group_role_uri UESURI of the group role where the role interface will be connected
        # @param [GroupAddRoleInterface] group DTO containing role interface that to be connected
        # @option group [String, UU::OS::UESURI] :role_interface_uri UESURI of the role interface that is to be connected to the specified group role. If nil is set, the command fails.
        #
        # @example
        #   # Add a role interface
        #   UU::OS::Group.add_role_interface('ues:TERRITORY:GROUP',
        #         :role_interface_uri => 'ues:TERRITORY:ROLE_INTERFACE'
        #   )
        def add_role_interface(group_role_uri, group = nil)
          svc = UU::OS::REST::RemoteClient.new(Group)
          payload = UU::OS::Group::GroupAddRoleInterface.new(group).to_json

          UU::OS::QoS::QoSHandler.auto_retry do
            svc.post('addRoleInterface', group_role_uri, payload)
          end
        end

        # Removes role interface connection from the specified group role.
        # Only connected interfaces can be disconnected from the group role. 
        #
        # @param [String, UU::OS::UESURI] group_role_uri UESURI of the group role where the role interface will be disconnected
        # @param [GroupRemoveRoleInterface] group DTO containing role interface that is to be disconnected
        # @option group [String, UU::OS::UESURI] :role_interface_uri UESURI of the role interface that is to be disconnected from the specified group role. If nil is set, the command fails.
        #
        # @example
        #   # Remove a role interface
        #   UU::OS::Group.remove_role_interface('ues:TERRITORY:GROUP',
        #         :role_interface_uri => 'ues:TERRITORY:ROLE_INTERFACE'
        #   )
        def remove_role_interface(group_role_uri, group = nil)
          svc = UU::OS::REST::RemoteClient.new(Group)
          payload = UU::OS::Group::GroupRemoveRoleInterface.new(group).to_json

          UU::OS::QoS::QoSHandler.auto_retry do
            svc.post('removeRoleInterface', group_role_uri, payload)
          end
        end

        # Returns list of the role interfaces that are connected to the specified group role.
        # The list of returned interfaces could be filtered by the name or code. Returned
        # list does not contain any objects filtered out by executed UESQuery. The list is
        # sorted by the name (and by code in case are equal) of an interface by default.
        #
        # @see https://uu.unicornuniverse.eu/ues/sesm?&SessFree=ues:UNI-BT:UES.DOC%2FCMD_V1%2FUESQUERY UESQuery Documentation.
        #
        # @param [String, UU::OS::UESURI] group_role_uri UESURI of the access role for which the connected role interfaces are to be listed
        # @param [GroupGetRoleInterfaceList] criteria DTO containing criteria for listing access role interfaces. Available parameters are: {GroupGetRoleInterfaceListItem#name name} and {GroupGetRoleInterfaceListItem#code code}.
        # @option criteria [String] :query The <b>UES query</b> string. It has to conform the UESQuery syntax (see documentation for more info). Filtering and order of the result list can be specified there. If nil is set, no filtering is applied and entries are ordered by name of casted subjects (and code as the second criteria).
        #
        # @return [UU::OS::REST::ResultList<Group::GroupGetRoleInterfaceList, Group::GroupGetRoleInterfaceListItem>] List of DTO representing role interfaces which are connected to the specified group role.
        #
        # @example
        #   # List all role interfaces
        #   UU::OS::Group.get_role_interface_list('ues:TERRITORY:GROUP')
        #
        #   # List role interfaces whose codes start with "IFC" prefix
        #   UU::OS::Group.get_role_interface_list('ues:TERRITORY:GROUP',
        #         :query => 'code="IFC%"'
        #   )
        def get_role_interface_list(group_role_uri, criteria = nil)
          svc = UU::OS::REST::RemoteClient.new(Group)
          dto = Group::GroupGetRoleInterfaceList.new(criteria)
          svc.add_parameter('pageIndex', dto.page_index)
          svc.add_parameter('pageSize', dto.page_size)
          svc.add_parameter('query', dto.query)
          UU::OS::QoS::QoSHandler.auto_retry do
            res = svc.get('getRoleInterfaceList', group_role_uri)
            return UU::OS::REST::ResultList.new(Group::GroupGetRoleInterfaceList, Group::GroupGetRoleInterfaceListItem, res)
          end
        end

      end
    end

end
