require 'uu/os/qos/qos_handler'
require 'uu/os/rest'
require 'uu/os/uesuri'
require 'uu/os/meta_role/meta_role_add_role_interface'
require 'uu/os/meta_role/meta_role_attributes'
require 'uu/os/meta_role/meta_role_copy'
require 'uu/os/meta_role/meta_role_create'
require 'uu/os/meta_role/meta_role_export'
require 'uu/os/meta_role/meta_role_get_export_data'
require 'uu/os/meta_role/meta_role_get_role_interface_list_item'
require 'uu/os/meta_role/meta_role_get_role_interface_list'
require 'uu/os/meta_role/meta_role_remove_role_interface'
require 'uu/os/meta_role/meta_role_set_attributes'

# Module Unicorn Universe.
module UU
  # Module Operating System.
  module OS
    # Module Meta Role.
    module MetaRole
      extend self

      # Service path
      PATH = 'ues/core/role/UESMetaRole'

      # Returns attributes representing the meta role specified by the meta_role_uri parameter.
      # The command does not change the state of the meta role in the system.
      #
      # @param [String, UU::OS::UESURI] meta_role_uri UESURI of a meta role whose attributes are returned
      #
      # @return [UU::OS::MetaRole::MetaRoleAttributes] DTO containing the meta role attributes
      #
      # @example
      #   # Get attributes of a meta role
      #   UU::OS::MetaRole.get_attributes('ues:TERRITORY:META_ROLE')
      def get_attributes(meta_role_uri)
        svc = UU::OS::REST::RemoteClient.new(MetaRole)
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('getAttributes', meta_role_uri)
          return MetaRole::MetaRoleAttributes.new(res)
        end
      end

      # Command for setting attributes of a meta role. The command cannot modify references to objects like
      # location, competent role and so on.
      #
      # @param [String, UU::OS::UESURI] meta_role_uri UESURI of a meta role whose attributes are returned
      # @param [UU::OS::MetaRole::MetaRoleSetAttributes] meta_role DTO containing new attributes of the meta role
      # @option meta_role [String] :name The new name of the meta role. If nil is set, the name is not modified.
      # @option meta_role [String] :code The new code of the meta role. If nil is set, the code is not modified. The code must be unique in a given territory.
      # @option meta_role [String] :description The new description of the meta role. If nil is set, the description is not modified.
      # @option meta_role [String, UU::OS::Artifact::SecurityLevel] :security_level The new security level of the meta role. If nil is set, the security level is not modified.
      # @option meta_role [String, UU::OS::UESURI] :icon_uri UESURI of the new meta role's icon. If nil is set, the icon is not modified. If nil UESURI ("ues:{[-1]}:{[-1]}:") is set, the meta role's icon is removed and the meta role uses icon from it's source meta artifact. An example of icon URI: 'ues:#{{}system}:#{{}ues_v5.core_v1.codetable_v1.base_v1.CodeTable_Icons}:#{{}ART_077}'.
      # @option meta_role [String] :default_name Specifies new default name of the created role according to meta role. If nil is set, the default name is not modified.
      # @option meta_role [String] :default_code Specifies new default code of the created role according to meta role. If nil is set, the default code is not modified.
      # @option meta_role [String, UU::OS::Artifact::SecurityLevel] :default_security_level Specifies new default security level of the created role according to meta role. If nil is set, the default security is not modified.
      # @option meta_role [String, UU::OS::UESURI] :default_icon_uri Specifies new default icon of the created role according to meta role (specified in meta_role_uri). If nil is set, the default icon is not modified. An example of icon URI: 'ues:#{{}system}:#{{}ues_v5.core_v1.codetable_v1.base_v1.CodeTable_Icons}:#{{}ART_077}'.
      # @option meta_role [Boolean] :default_searchable_in_territory Specifies if it is possible to search for the role created according to the meta role in its territory. If nil is set, the attribute is not modified.
      # @option meta_role [Boolean] :component_content Defines whether the primarily displayed content of the created Meta role is component content. Default value is false.           
      # @option meta_role [Boolean] :default_component_content Defines whether the primarily displayed content of a role created according to the meta role (specified in source_meta_artifact_uri) is component content. When nil is set, the default component content setting is not modified.
      #
      # @return [UU::OS::UESURI] UESURI of the meta role
      #
      # @example
      #   # Set attributes of a meta role
      #   UU::OS::MetaRole.set_attributes('ues:TERRITORY:META_ROLE', :name => 'New Meta Role', :code => 'NEW_META_ROLE',
      #     :description => 'Meta Role designated for testing purposes', :default_searchable_in_territory => true)
      def set_attributes(meta_role_uri, meta_role = nil)
        svc = UU::OS::REST::RemoteClient.new(MetaRole)
        payload = UU::OS::MetaRole::MetaRoleSetAttributes.new(meta_role).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.post('setAttributes', meta_role_uri, payload)
          return UU::OS::UESURI.new(res)
        end
      end

      # Copy a meta role. The command creates a new meta role according to specified source
      # meta role in a specified location. At the very least location meta role and name must
      # be specified in {MetaRoleCopy} DTO. Competent role for the new meta role copy will
      # be selected as the most suitable according to specified container and meta role
      # (executive/authorized role with connected interface), or can be also specified in {MetaRoleCopy} DTO.
      #
      # @param [String, UU::OS::UESURI] source_meta_artifact_uri UESURI of the meta role which will be copied.
      # @param [UU::OS::MetaRole::MetaRoleCopy] meta_role DTO containing attributes of the new meta role copy.
      # @option meta_role [String, UU::OS::UESURI] :location_uri UESURI of the meta model dictionary where to the meta role will be copied.
      # @option meta_role [String] :name Name of the new meta role copy. The value must not be blank.
      # @option meta_role [String] :code Code of the new meta role copy.
      # @option meta_role [String] :description Description of the new meta role copy.
      # @option meta_role [String, UU::OS::UESURI] :competent_role_uri UESURI of the competent role.
      # @option meta_role [String, UU::OS::Artifact::SecurityLevel] :security_level Security level.
      # @option meta_role [String, UU::OS::UESURI] :icon_uri URI of the new meta role's icon. Allowed format of icon URI: 'ues:#{{}system}:#{{}ues_v5.core_v1.codetable_v1.base_v1.CodeTable_Icons}:#{{}CODE_OF_ICON}'.
      # @option meta_role [Boolean] :copy_sheets Specifies whether to copy content from the source meta role to the new copy. Default value is false.
      # @option meta_role [Boolean] :copy_lifecycle Specifies whether to copy workflow from the source meta role to the new copy. Default value is false.
      # @option meta_role [Boolean] :copy_attachments Specifies whether to copy attachments from the source meta role to the new copy. Default value is false.
      # @option meta_role [Boolean] :copy_sheet_templates Specifies whether to copy sheet templates from the source meta role to the new copy. Default value is false.
      # @option meta_role [String] :default_name Specifies default name of the created role according to new meta role copy.
      # @option meta_role [String] :default_code Specifies default code of the created role according to new meta role copy.
      # @option meta_role [String, UU::OS::Artifact::SecurityLevel] :default_security_level Specifies default security level of the created role according to new meta role copy.
      # @option meta_role [String, UU::OS::UESURI] :default_icon_uri Specifies default icon of the created role according to new meta role copy. Allowed format of icon URI: 'ues:#{{}system}:#{{}ues_v5.core_v1.codetable_v1.base_v1.CodeTable_Icons}:#{{}CODE_OF_ICON}'.
      # @option meta_role [String] :default_searchable_in_territory Specifies if it is possible to search for the role created according to the meta role in its territory.
      # @option meta_role [Boolean] :component_content Defines whether the primarily displayed content of the created Meta role is component content. Default value is false.           
      # @option meta_role [Boolean] :default_component_content Defines whether the primarily displayed content of a role created according to the new meta role copy is component content. When nil is set, the default component content setting is copied from the source meta artifact.
      #
      # @return [UU::OS::UESURI] UESURI of the copied meta role.
      #
      # @example
      #   # Copy a meta role (including its sheets); change a few attributes
      #   UU::OS::MetaRole.copy('ues:TERRITORY:META_ROLE', :location_uri => 'ues:TERRITORY:APP_PACKAGE', :code => 'META_ROLE_COPY',
      #     :icon_uri => 'ues:#{system}:#{ues_v5.core_v1.codetable_v1.base_v1.CodeTable_Icons}:#{ART_077}',
      #     :security_level => UU::OS::Artifact::SecurityLevel::INTERNAL_USE, :copy_sheets => true)
      def copy(source_meta_artifact_uri, meta_role = nil)
        svc = UU::OS::REST::RemoteClient.new(MetaRole)
        svc.timeout = 120 # may take longer than default 60 seconds
        payload = UU::OS::MetaRole::MetaRoleCopy.new(meta_role).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.post('copy', source_meta_artifact_uri, payload)
          return UU::OS::UESURI.new(res)
        end
      end

      # Creates new role interface connection to the specified meta role.
      # The role interface has to be in an active state and must not be already added. The role
      # interface can be added to meta role only provided that role interface is located in the
      # same, child or parent meta model as meta role.
      #
      # @param [String, UU::OS::UESURI] meta_role_uri UESURI of the role where the role interface will be connected
      # @param [UU::OS::MetaRole::MetaRoleAddRoleInterface] role_interface DTO containing role interface that is to be connected
      # @option role_interface [String, UU::OS::UESURI] :role_interface_uri UESURI of the role interface that is to be connected to the specified meta role.
      #
      # @example
      #   # Connect another role interface to a meta role
      #   UU::OS::MetaRole.add_role_interface('ues:TERRITORY:META_ROLE', :role_interface_uri => 'ues:TERRITORY:ROLE_INTERFACE')
      def add_role_interface(meta_role_uri, role_interface = nil)
        svc = UU::OS::REST::RemoteClient.new(MetaRole)
        payload = UU::OS::MetaRole::MetaRoleAddRoleInterface.new(role_interface).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          svc.post('addRoleInterface', meta_role_uri, payload)
        end
      end

      # Returns list of the role interfaces that are added to the specified meta role.
      # The list of returned interfaces could be filtered by the name or code. Returned list does
      # not contain any objects filtered out by executed UESQuery. The list is sorted by the name
      # (and codes in case are equal) of an interface by default.
      #
      # See {https://uu.unicornuniverse.eu/ues/sesm?&SessFree=ues:UNI-BT:UES.DOC%2FCMD_V1%2FUESQUERY UESQuery Documentation}.
      #
      # @param [String, UU::OS::UESURI] meta_role_uri UESURI of the meta role for which the added role interfaces are to be listed
      # @param [UU::OS::MetaRole::MetaRoleGetRoleInterfaceList] criteria DTO containing criteria for listing connected role interfaces
      # @option criteria [String] :query Query for filtering and ordering of the result list (more information in {uuQuery - Guideline}[https://plus4u.net/ues/sesm?SessFree=ues%253AUNI-BT%255B210795%255D%253AUES.DOC%252FCMD_V1%252FUESQUERY%255B44191574519623267%255D%253A]). It is possible to filter the result list by the name or code.
      #
      # @return [UU::OS::REST::ResultList<MetaRole::MetaRoleGetRoleInterfaceList, MetaRole::MetaRoleGetRoleInterfaceListItem>] List of DTO representing role interfaces which are connected to the specified meta role
      #
      # @example
      #   # Get a list of all role interfaces which are connected to the specified meta role
      #   UU::OS::MetaRole.get_role_interface_list('ues:TERRITORY:META_ROLE')
      #
      #   # Get a list of role interfaces with name that starts with string John.
      #   UU::OS::MetaRole.get_role_interface_list('ues:TERRITORY:META_ROLE', :query => "name='John%'")
      def get_role_interface_list(meta_role_uri, criteria = nil)
        svc = UU::OS::REST::RemoteClient.new(MetaRole)
        dto = MetaRole::MetaRoleGetRoleInterfaceList.new(criteria)
        svc.add_parameter('pageIndex', dto.page_index)
        svc.add_parameter('pageSize', dto.page_size)
        svc.add_parameter('query', dto.query)
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('getRoleInterfaceList', meta_role_uri)
          return UU::OS::REST::ResultList.new(MetaRole::MetaRoleGetRoleInterfaceList, MetaRole::MetaRoleGetRoleInterfaceListItem, res)
        end
      end

      # Removes role interface connection from the specified meta role.
      # Only added interfaces can be removed from meta role.
      #
      # @param [String, UU::OS::UESURI] meta_role_uri UESURI of the meta role where the role interface will be disconnected.
      # @param [UU::OS::MetaRole::MetaRoleRemoveRoleInterface] role_interface DTO containing role interface that to be disconnected
      # @option role_interface [String, UU::OS::UESURI] :role_interface_uri UESURI of the role interface that to be removed from the specified meta role.
      #
      # @example
      #   # Remove a connected role interface from a meta role
      #   UU::OS::MetaRole.remove_role_interface('ues:TERRITORY:META_ROLE', :role_interface_uri => 'ues:TERRITORY:ROLE_INTERFACE')
      def remove_role_interface(meta_role_uri, role_interface = nil)
        svc = UU::OS::REST::RemoteClient.new(MetaRole)
        payload = UU::OS::MetaRole::MetaRoleRemoveRoleInterface.new(role_interface).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          svc.post('removeRoleInterface', meta_role_uri, payload)
        end
      end

      # Exports a meta role to the XML file, which is saved to the export storage. The export is an asynchronous process.<br>
      # It is possible to wait for the end of the process. More information in {UU::OS::Env::Process Process} and in {UU::OS::REST::Future Future}.<br>
      # To get exported data from the storage use {#get_export_data get_export_data}.
      #
      # @param [String, UU::OS::UESURI] meta_role_uri UESURI of the meta role to export.
      #
      # @return [UU::OS::UESURI] UESURI of the export process.
      #
      # @example
      #   # Start exporting a meta role.
      #   process_uri = UU::OS::MetaRole.export('ues:TERRITORY:META_ROLE')
      #
      #   # Wait for the end of the process.
      #   export_uri = UU::OS::REST::Future.new(process_uri).get(nil, 600)
      #
      #   # Get exported XML file.
      #   exported_data = UU::OS::MetaRole.get_export_data(export_uri)
      def export(meta_role_uri)
        svc = UU::OS::REST::RemoteClient.new(MetaRole)
        payload = UU::OS::MetaRole::MetaRoleExport.new(nil).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.post('export', meta_role_uri, payload)
          return UU::OS::UESURI.new(res)
        end
      end

      # Returns a XML file generated by the export command.
      #
      # @param [String, UU::OS::UESURI] meta_role_uri UESURI of the meta artifact we want to get the exported XML file from.
      # @param [UU::OS::MetaRole::MetaRoleGetExportData] meta_role Specific options.
      # @option meta_artifact [String] :requested_mime_type MIME type of the returned value. Possible values are 'application/xml' for XML and 'application/zip' for ZIP. ZIP is returned by default.
      #
      # @return [UU::OS::REST::BinaryValue] Data file of an exported meta artifact (XML or ZIP containing XML).
      #
      # @example
      #   # Start exporting a meta role.
      #   process_uri = UU::OS::MetaRole.export('ues:TERRITORY:META_ROLE')
      #
      #   # Wait for the end of the process.
      #   export_uri = UU::OS::REST::Future.new(process_uri).get(nil, 600)
      #
      #   # Get exported XML file.
      #   exported_data = UU::OS::MetaRole.get_export_data(export_uri)
      def get_export_data(meta_role_uri, meta_role = nil)
        svc = UU::OS::REST::RemoteClient.new(MetaRole)
        dto = UU::OS::MetaRole::MetaRoleGetExportData.new(meta_role)
        if dto.requested_mime_type
          svc.add_parameter('requestedMimeType', dto.requested_mime_type)
        end
        UU::OS::QoS::QoSHandler.auto_retry do
          return svc.get_binary('getExportData', meta_role_uri)
        end
      end

      # Creates a meta role from an exported XML file in the specified location. Source meta role must be derived (at least indirectly) from a meta artifact which is a part of UU.OS uuApp. Creation of meta role is divided into synchronous and asynchronous part. During the synchronous part, the meta role is created. During the asynchronous part, the template objects are created. If the asynchronous process fails and the Problem state is set on meta role, it is possible to solve the problem (e.g. add reference_map) and run the command again.
      # @deprecated The uuCommand MetaRole.create is used for creating meta role from the XML file temporarily. This feature will be removed in a future version and replaced by the specific uuCommands for creating template objects.
      #
      # @param [String, UU::OS::UESURI] location_uri UESURI of the meta model dictionary where new meta role should be created.
      # @param [UU::OS::MetaRole::MetaRoleCreate] meta_role Specific options.
      # @option meta_role [String, File, UU::OS::REST::BinaryValue] :data XML or zip compressed XML containing informations about meta role (e.g. exported with {export})
      # @option meta_role [String, UU::OS::UESURI] :competent_role_uri UESURI of the role that will be competent for the new meta role. If nil is set, the attribute from the XML (entered in :data option) is used.
      # @option meta_role [String] :name Name of the new meta role. If nil is set, the attribute from the XML (entered in :data option) is used.
      # @option meta_role [String] :code Code of the new meta role. If nil is set, the attribute from the XML (entered in :data option) is used.
      # @option meta_role [String] :description Description of the new meta role. If nil is set, the attribute from the XML (entered in :data option) is used.
      # @option meta_role [String, UU::OS::Artifact::SecurityLevel] :security_level Security level of the new meta role. If nil is set, the attribute from the XML (entered in :data option) is used.
      # @option meta_role [String, UU::OS::UESURI] :icon_uri UESURI of the new meta role icon. If nil is set, the attribute from the XML (entered in :data option) is used. An example of icon URI: 'ues:#{{}system}:#{{}ues_v5.core_v1.codetable_v1.base_v1.CodeTable_Icons}:#{{}ART_077}'.
      # @option meta_role [String] :default_name Default name of a role that will be created according to the new meta role. If nil is set, the attribute from the XML (entered in :data option) is used.
      # @option meta_role [String] :default_code Default code of a role that will be created according to the new meta role. If nil is set, the attribute from the XML (entered in :data option) is used.
      # @option meta_role [String, UU::OS::UESURI] :default_icon_uri UESURI of the artifact icon that will be created according to the new meta role. If nil is set, the attribute from the XML (entered in :data option) is used. An example of icon URI: 'ues:#{{}system}:#{{}ues_v5.core_v1.codetable_v1.base_v1.CodeTable_Icons}:#{{}ART_077}'.
      # @option meta_role [String, UU::OS::Artifact::SecurityLevel] :default_security_level Default security level of the new meta role. If nil is set, the attribute from the XML (entered in :data option) is used.
      # @option meta_role [Boolean] :default_searchable_in_territory Flag indicating whether a role that will be created according to the new meta role is searchable in the territory.
      # @option meta_role [Boolean] :import_life_cycle_template Flag indicating whether the life cycle template will be created according to specification in the entered XML. Default value is true.
      # @option meta_role [Boolean] :import_property_templates Flag indicating whether the property templates will be created according to specification in the entered XML. Default value is true.
      # @option meta_role [Boolean] :import_sheet_templates Flag indicating whether the sheet templates will be created according to specification in the entered XML. Default value is true.
      # @option meta_role [Boolean] :import_properties Flag indicating whether the properties will be created according to specification in the entered XML. Default value is true.
      # @option meta_role [Boolean] :import_sheets Flag indicating whether the sheets will be created according to specification in the entered XML. Default value is true.
      # @option meta_role [Boolean] :import_attachments Flag indicating whether the attachments will be created according to specification in the entered XML. Default value is true.      
      # @option meta_role [Boolean] :import_meta_permissions Flag indicating whether the meta permissions will be created according to specification in the entered XML. Default value is true.
      # @option meta_role [Boolean] :import_permissions Flag indicating whether the permissions will be created according to specification in the entered XML. Default value is true.
      # @option meta_role [Boolean] :import_use_cases Flag indicating whether the use cases will be created according to specification in the entered XML. Default value is true.
      # @option meta_role [Boolean] :import_masks Flag indicating whether the permission masks will be created according to specification in the entered XML. Default value is true.
      # @option meta_role [String] :reference_map JSON describing the linkage of references specified in XML to another objects (e.g. references in properties, activity related artifacts. ...).
      # @option meta_role [Boolean] :component_content Defines whether the primarily displayed content of the created Meta role is component content. Default value is false.           
      # @option meta_role [Boolean] :default_component_content Defines whether the primarily displayed content of a role created according to the new meta role is component content. When nil is set, the attribute from the XML (entered in :data option) is used.
      #
      # @return [UU::OS::UESURI] UESURI of the new meta role.
      #
      # @example
      #   # Get exported data from an existing meta role.
      #   exported_data = UU::OS::MetaRole.get_export_data('ues:TERRITORY:META_ROLE', :requested_mime_type => 'application/xml')
      #
      #   # Create a new meta role according to exported one without creating sheets and attachments specified in the XML. Also set different code and name.
      #   UU::OS::MetaRole.create('ues:TERRITORY:APP_PACKAGE', :code => 'NEW_META_ROLE', :name => 'New Meta Role', :data => exported_data.data, :import_sheets => false, :import_attachments => false)
      #
      #   # Create a new meta role and set new objects to the references.
      #   UU::OS::MetaRole.create('ues:TERRITORY:APP_PACKAGE', :data => exported_data.data, :reference_map => '[{"ues:OLD_TERRITORY:OLD_ARTIFACT1":"ues:TERRITORY:ARTIFACT1"},{"ues:OLD_TERRITORY:OLD_ARTIFACT2:OLD_OBJECT":"ues:TERRITORY:ARTIFACT2:OBJECT"}]')
      def create(location_uri, meta_role = nil)
        svc = UU::OS::REST::RemoteClient.new(MetaRole)
        dto = UU::OS::MetaRole::MetaRoleCreate.new(meta_role)
        payload = dto.to_hash
        payload.delete(:data)
        tmpfile = Tempfile.new("MetaRoleCreate")
        begin
          if (!dto.data.data.respond_to?(:read))
            tmpfile.write(dto.data.data)
            payload[:data] = tmpfile
          elsif (dto.data.data.respond_to?(:read) && !dto.data.data.respond_to?(:path))
            tmpfile.binmode
            dto.data.data.rewind
            tmpfile.write(dto.data.data.read)
            payload[:data] = tmpfile
          else
            payload[:data] = dto.data.to_hash(false)
          end
          tmpfile.flush
          tmpfile.rewind
              
          UU::OS::QoS::QoSHandler.auto_retry do
            res = svc.post('create', location_uri, payload)
            return UU::OS::UESURI.new(res)
          end
        ensure
          tmpfile.close
          tmpfile.unlink
        end
      end

    end
  end

end
